// SPDX-License-Identifier: LGPL-3.0-or-later
// See Notices.txt for copyright information

use crate::{
    expr::Expr,
    intern::{Intern, Interned},
    source_location::SourceLocation,
    ty::{impl_match_variant_as_self, CanonicalType, StaticType, Type, TypeProperties},
    util::{ConstBool, ConstUsize, GenericConstBool},
};
use bitvec::slice::BitSlice;
use std::{
    borrow::BorrowMut,
    marker::PhantomData,
    ops::{Bound, Index, Range, RangeBounds},
    sync::Arc,
};

mod sealed {
    pub trait IntTypeSealed {}
    pub trait SizeSealed {}
}

pub trait Size:
    sealed::SizeSealed + Copy + Ord + std::hash::Hash + std::fmt::Debug + Send + Sync + 'static
{
    type ArrayMatch<Element: Type>: AsRef<[Expr<Element>]>
        + AsMut<[Expr<Element>]>
        + BorrowMut<[Expr<Element>]>
        + 'static
        + Send
        + Sync
        + Eq
        + Clone
        + std::hash::Hash
        + std::fmt::Debug
        + IntoIterator<Item = Expr<Element>>
        + TryFrom<Vec<Expr<Element>>>
        + Into<Vec<Expr<Element>>>;
    const KNOWN_VALUE: Option<usize>;
    fn as_usize(self) -> usize;
    fn try_from_usize(v: usize) -> Option<Self>;
    #[track_caller]
    fn from_usize(v: usize) -> Self {
        Self::try_from_usize(v).expect("wrong size")
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Hash)]
#[repr(transparent)]
pub struct DynSize(pub usize);

impl std::fmt::Debug for DynSize {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        self.0.fmt(f)
    }
}

impl sealed::SizeSealed for DynSize {}

impl Size for DynSize {
    type ArrayMatch<Element: Type> = Box<[Expr<Element>]>;

    const KNOWN_VALUE: Option<usize> = None;

    fn as_usize(self) -> usize {
        self.0
    }

    fn try_from_usize(v: usize) -> Option<Self> {
        Some(DynSize(v))
    }
}

impl<const VALUE: usize> sealed::SizeSealed for ConstUsize<VALUE> {}

impl<const VALUE: usize> Size for ConstUsize<VALUE> {
    type ArrayMatch<Element: Type> = [Expr<Element>; VALUE];

    const KNOWN_VALUE: Option<usize> = Some(VALUE);

    fn as_usize(self) -> usize {
        VALUE
    }

    fn try_from_usize(v: usize) -> Option<Self> {
        if v == VALUE {
            Some(Self)
        } else {
            None
        }
    }
}

pub trait KnownSize: crate::util::GenericConstUsize + Size + Default {}

impl<const VALUE: usize> KnownSize for ConstUsize<VALUE> {}

macro_rules! impl_int {
    ($name:ident, $generic_name:ident, $value:ident, $SIGNED:literal) => {
        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
        #[repr(transparent)]
        pub struct $name<Width: Size = DynSize> {
            pub width: Width,
        }

        #[allow(non_upper_case_globals)]
        pub const $name: $generic_name = $generic_name;

        impl<Width: Size> $name<Width> {
            pub fn new(width: Width) -> Self {
                Self { width }
            }
            pub fn width(self) -> usize {
                self.width.as_usize()
            }
            pub fn type_properties(self) -> TypeProperties {
                TypeProperties {
                    is_passive: true,
                    is_storable: true,
                    is_castable_from_bits: true,
                    bit_width: self.width(),
                }
            }
        }

        impl<Width: Size> sealed::IntTypeSealed for $name<Width> {}

        impl<Width: Size> IntType for $name<Width> {
            type Dyn = $name;
            type Width = Width;
            type Signed = ConstBool<$SIGNED>;
            fn width(self) -> usize {
                $name::width(self)
            }
            fn new(width: Self::Width) -> Self {
                $name { width }
            }
            fn as_same_width_sint(self) -> SInt<Self::Width> {
                SInt { width: self.width }
            }
            fn as_same_width_uint(self) -> UInt<Self::Width> {
                UInt { width: self.width }
            }
        }

        impl $name {
            pub fn new_dyn(width: usize) -> Self {
                Self {
                    width: DynSize(width),
                }
            }
        }

        impl<Width: KnownSize> $name<Width> {
            pub fn new_static() -> Self {
                Self {
                    width: Width::default(),
                }
            }
        }

        impl<Width: Size> Type for $name<Width> {
            type MaskType = Bool;
            impl_match_variant_as_self!();
            fn mask_type(&self) -> Self::MaskType {
                Bool
            }
            fn canonical(&self) -> CanonicalType {
                CanonicalType::$name($name::new_dyn(self.width()))
            }
            #[track_caller]
            fn from_canonical(canonical_type: CanonicalType) -> Self {
                let CanonicalType::$name(retval) = canonical_type else {
                    panic!("expected {}", stringify!($name));
                };
                $name {
                    width: Width::from_usize(retval.width.as_usize()),
                }
            }
            fn source_location() -> SourceLocation {
                SourceLocation::builtin()
            }
        }

        impl<Width: KnownSize> StaticType for $name<Width> {
            fn static_type() -> Self {
                Self::new_static()
            }
        }

        #[derive(Copy, Clone, PartialEq, Eq, Hash, Debug, Default)]
        pub struct $generic_name;

        impl Index<usize> for $generic_name {
            type Output = $name;

            fn index(&self, width: usize) -> &Self::Output {
                Interned::<_>::into_inner(Intern::intern_sized($name::new_dyn(width)))
            }
        }

        #[derive(Clone, PartialEq, Eq, Hash, Debug)]
        pub struct $value<Width: Size = DynSize> {
            bits: Arc<BitSlice>,
            _phantom: PhantomData<Width>,
        }

        impl<Width: Size> $value<Width> {
            pub fn width(&self) -> usize {
                if let Some(retval) = Width::KNOWN_VALUE {
                    debug_assert!(self.bits.len() == retval);
                    retval
                } else {
                    self.bits.len()
                }
            }
            pub fn ty(&self) -> $name<Width> {
                $name {
                    width: Width::from_usize(self.width()),
                }
            }
            pub fn as_dyn_int(self) -> $value {
                $value {
                    bits: self.bits,
                    _phantom: PhantomData,
                }
            }
        }

        impl<Width: KnownSize> $value<Width> {
            pub fn new_static(bits: Arc<BitSlice>) -> Self {
                assert!(bits.len() == Width::VALUE);
                Self {
                    bits,
                    _phantom: PhantomData,
                }
            }
        }

        impl $value {
            pub fn new_dyn(bits: Arc<BitSlice>) -> Self {
                Self {
                    bits,
                    _phantom: PhantomData,
                }
            }
        }
    };
}

impl_int!(UInt, UIntWithoutGenerics, UIntValue, false);
impl_int!(SInt, SIntWithoutGenerics, SIntValue, true);

pub trait IntType: Type + sealed::IntTypeSealed {
    type Dyn: IntType<Dyn = Self::Dyn, Signed = Self::Signed, Width = DynSize>;
    type Signed: GenericConstBool;
    type Width: Size;
    fn width(self) -> usize;
    fn as_dyn_int(self) -> Self::Dyn {
        Self::new_dyn(self.width())
    }
    fn new_dyn(width: usize) -> Self::Dyn {
        Self::Dyn::new(DynSize(width))
    }
    fn new(width: Self::Width) -> Self;
    fn new_static() -> Self
    where
        Self::Width: KnownSize,
    {
        Self::new(Self::Width::default())
    }
    fn as_same_width_sint(self) -> SInt<Self::Width>;
    fn as_same_width_uint(self) -> UInt<Self::Width>;
    fn slice_index_to_range<I: RangeBounds<usize>>(self, index: I) -> Range<usize> {
        let width = self.width();
        let start = match index.start_bound() {
            Bound::Included(start) => *start,
            Bound::Excluded(start) => *start + 1,
            Bound::Unbounded => 0,
        };
        let end = match index.end_bound() {
            Bound::Included(end) => *end + 1,
            Bound::Excluded(end) => *end,
            Bound::Unbounded => width,
        };
        assert!(start <= end && end <= width, "slice range out-of-range");
        start..end
    }
    fn slice_and_shift<I: RangeBounds<usize>>(self, index: I) -> (UInt, usize) {
        let range = self.slice_index_to_range(index);
        let width = range.end - range.start;
        (UInt::new_dyn(width), range.start)
    }
    fn slice<I: RangeBounds<usize>>(self, index: I) -> UInt {
        self.slice_and_shift(index).0
    }
}

#[derive(Copy, Clone, PartialEq, Eq, Hash, Debug)]
pub struct Bool;

impl Bool {
    pub fn type_properties(self) -> TypeProperties {
        TypeProperties {
            is_passive: true,
            is_storable: true,
            is_castable_from_bits: true,
            bit_width: 1,
        }
    }
}

impl Type for Bool {
    type MaskType = Bool;
    impl_match_variant_as_self!();
    fn mask_type(&self) -> Self::MaskType {
        Bool
    }
    fn canonical(&self) -> CanonicalType {
        CanonicalType::Bool(Bool)
    }
    #[track_caller]
    fn from_canonical(canonical_type: CanonicalType) -> Self {
        let CanonicalType::Bool(retval) = canonical_type else {
            panic!("expected Bool");
        };
        retval
    }
    fn source_location() -> SourceLocation {
        SourceLocation::builtin()
    }
}

impl StaticType for Bool {
    fn static_type() -> Self {
        Bool
    }
}

pub trait IntCmp<Rhs> {
    fn cmp_eq(self, rhs: Rhs) -> Expr<Bool>;
    fn cmp_ne(self, rhs: Rhs) -> Expr<Bool>;
    fn cmp_lt(self, rhs: Rhs) -> Expr<Bool>;
    fn cmp_le(self, rhs: Rhs) -> Expr<Bool>;
    fn cmp_gt(self, rhs: Rhs) -> Expr<Bool>;
    fn cmp_ge(self, rhs: Rhs) -> Expr<Bool>;
}
